'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var compilerSfc = require('@vue/compiler-sfc');
var fs = _interopDefault(require('fs-extra'));
var createDebugger = _interopDefault(require('debug'));
var path = require('path');
var pluginutils = require('@rollup/pluginutils');
var qs = _interopDefault(require('query-string'));
var crypto = require('crypto');

const hash = data => crypto.createHash("sha256").update(data).digest("hex").slice(0, 8);
const joinCode = code => `${code.join("\n")}\n`;
function normalizePath(...paths) {
  const f = path.join(...paths).replace(/\\/g, "/");
  if (/^\.[/\\]/.test(paths[0])) return `./${f}`;
  return f;
}

try {
  require.resolve("@vue/compiler-sfc");
} catch {
  throw new Error("rollup-plugin-vue3 requires @vue/compiler-sfc");
}
const debug = createDebugger("rollup-plugin-vue3");
const cache = new Map();
const defaultOpts = {
  include: /\.vue$/,
  exclude: [],
  target: "browser",
  exposeFilename: false,
  customBlocks: []
};
var index = ((opts = {}) => {
  const options = { ...defaultOpts,
    ...opts
  };
  const isServer = options.target === "node";
  const isProduction = "production" === "production" ;
  const rootContext = process.cwd();
  const filter = pluginutils.createFilter(options.include, options.exclude);
  const filterCustomBlock = createCustomBlockFilter(options.customBlocks);
  const plugin = {
    name: "vue3",

    async resolveId(id, importer) {
      const query = parseQuery(id);
      if (!query.vue) return null;

      if (query.src) {
        const resolved = await this.resolve(query.filename, importer, {
          skipSelf: true
        });

        if (resolved && importer) {
          cache.set(resolved.id, getDescriptor(importer));
          const [, originalQuery] = id.split("?", 2);
          resolved.id += `?${originalQuery}`;
          return resolved;
        }
      } else if (!filter(query.filename)) return null;

      debug(`resolveId(${id})`);
      return id;
    },

    async load(id) {
      const query = parseQuery(id);
      if (!query.vue) return null;
      if (query.src) return fs.readFile(query.filename, "utf-8");
      const descriptor = getDescriptor(query.filename);

      if (descriptor) {
        let block = null;
        if (query.type === "template") block = descriptor.template;else if (query.type === "script") block = descriptor.script;else if (query.type === "style") block = descriptor.styles[query.index];else if (query.type === "custom") block = descriptor.customBlocks[query.index];
        if (block) return {
          code: block.content,
          map: normalizeSourceMap(block.map)
        };
      }

      return null;
    },

    async transform(code, id) {
      const query = parseQuery(id); // .vue file

      if (!query.vue && filter(id)) {
        debug(`transform(${id})`);
        const {
          descriptor,
          errors
        } = parseSFC(code, id, rootContext);

        if (errors.length > 0) {
          for (const error of errors) this.error(createRollupError(id, error));

          return null;
        }

        const output = transformVueSFC(code, id, descriptor, {
          rootContext,
          isProduction,
          isServer,
          filterCustomBlock
        }, options);
        debug("transient .vue file:", `\n${output}\n`);
        return {
          code: output,
          map: {
            mappings: ""
          }
        };
      }

      if (!query.vue) return null;
      if (!query.src && !filter(query.filename)) return null;
      const descriptor = getDescriptor(query.filename);
      const hasScoped = descriptor.styles.some(s => s.scoped);

      if (query.type === "template" && descriptor.template) {
        debug(`transform(${id})`);
        const block = descriptor.template;
        const preprocessLang = block.lang;
        const preprocessOptions = preprocessLang && options.templatePreprocessOptions && options.templatePreprocessOptions[preprocessLang];
        const result = compilerSfc.compileTemplate({
          filename: query.filename,
          source: code,
          inMap: query.src ? undefined : block.map,
          preprocessLang,
          preprocessOptions,
          preprocessCustomRequire: options.preprocessCustomRequire,
          compiler: options.compiler,
          ssr: isServer,
          compilerOptions: { ...options.compilerOptions,
            scopeId: hasScoped && query.id ? `data-v-${query.id}` : undefined,
            bindingMetadata: descriptor.script ? descriptor.script.bindings : undefined
          },
          transformAssetUrls: options.transformAssetUrls
        });

        if (result.errors.length > 0) {
          for (const error of result.errors) this.error(typeof error === "string" ? {
            id: query.filename,
            message: error
          } : createRollupError(query.filename, error));

          return null;
        }

        if (result.tips.length > 0) for (const tip of result.tips) this.warn({
          id: query.filename,
          message: tip
        });
        return {
          code: result.code,
          map: normalizeSourceMap(result.map)
        };
      } else if (query.type === "style") {
        debug(`transform(${id})`);
        const block = descriptor.styles[query.index];
        let preprocessOptions = options.preprocessOptions || {};
        const preprocessLang = options.preprocessStyles ? block.lang : undefined;

        if (preprocessLang) {
          preprocessOptions = preprocessOptions[preprocessLang] || preprocessOptions; // include node_modules for imports by default

          switch (preprocessLang) {
            case "scss":
            case "sass":
              preprocessOptions = {
                includePaths: ["node_modules"],
                ...preprocessOptions
              };
              break;

            case "less":
            case "stylus":
              preprocessOptions = {
                paths: ["node_modules"],
                ...preprocessOptions
              };
          }
        } else preprocessOptions = {};

        const result = await compilerSfc.compileStyleAsync({
          filename: query.filename,
          id: query.id ? `data-v-${query.id}` : ``,
          source: code,
          scoped: block.scoped,
          vars: !!block.vars,
          modules: !!block.module,
          postcssOptions: options.postcssOptions,
          postcssPlugins: options.postcssPlugins,
          modulesOptions: options.cssModulesOptions,
          preprocessLang,
          preprocessCustomRequire: options.preprocessCustomRequire,
          preprocessOptions
        });

        if (result.errors.length > 0) {
          for (const error of result.errors) this.error({
            id: query.filename,
            message: error.message
          });

          return null;
        }

        if (query.module) return {
          code: `export default ${JSON.stringify(result.modules)};`,
          map: null
        };
        return {
          code: result.code,
          map: normalizeSourceMap(result.map)
        };
      }

      return null;
    }

  };
  return plugin;
});

function createCustomBlockFilter(queries) {
  if (!queries || queries.length === 0) return () => false;
  const allowed = new Set(queries.filter(query => /^[a-z]/i.test(query)));
  const disallowed = new Set(queries.filter(query => /^![a-z]/i.test(query)).map(query => query.slice(1)));
  const allowAll = queries.includes("*") || !queries.includes("!*");
  return type => {
    if (allowed.has(type)) return true;
    if (disallowed.has(type)) return true;
    return allowAll;
  };
}

function parseQuery(id) {
  const [filename, query] = id.split("?", 2);
  if (!query) return {
    vue: false
  };
  const raw = qs.parse(query);
  if (!("vue" in raw)) return {
    vue: false
  };
  return { ...raw,
    filename,
    vue: true,
    index: raw.index && Number(raw.index),
    src: "src" in raw,
    scoped: "scoped" in raw
  };
}

function getDescriptor(id) {
  const cached = cache.get(id);
  if (cached) return cached;
  throw new Error(`${id} is not parsed yet`);
}

function parseSFC(code, id, sourceRoot) {
  const {
    descriptor,
    errors
  } = compilerSfc.parse(code, {
    sourceMap: true,
    filename: id,
    sourceRoot: sourceRoot
  });
  cache.set(id, descriptor);
  return {
    descriptor,
    errors: errors
  };
}

function transformVueSFC(code, resourcePath, descriptor, {
  rootContext,
  isProduction,
  isServer,
  filterCustomBlock
}, options) {
  const shortFilePath = path.relative(rootContext, resourcePath).replace(/^(\.\.[/\\])+/, "").replace(/\\/g, "/");
  const id = hash(isProduction ? `${shortFilePath}\n${code}` : shortFilePath);
  const hasScoped = descriptor.styles.some(s => s.scoped);
  const templateImport = getTemplateCode(descriptor, resourcePath, id, hasScoped, isServer);
  const scriptImport = getScriptCode(descriptor, resourcePath);
  const stylesCode = getStyleCode(descriptor, resourcePath, id, options.preprocessStyles);
  const customBlocksCode = getCustomBlock(descriptor, resourcePath, filterCustomBlock);
  const output = [scriptImport, templateImport, stylesCode, customBlocksCode, isServer ? `script.ssrRender = ssrRender;` : `script.render = render;`];
  if (hasScoped) output.push(`script.__scopeId = "data-v-${id}";`);

  if (!isProduction) {
    output.push(`script.__file = "${shortFilePath}";`);
  } else if (options.exposeFilename) {
    output.push(`script.__file = "${path.basename(shortFilePath)}";`);
  }

  output.push("export default script");
  return joinCode(output);
}

function getTemplateCode(descriptor, resourcePath, id, hasScoped, isServer) {
  let templateImport = `const render = () => {};`;

  if (descriptor.template) {
    var _descriptor$template$;

    const src = (_descriptor$template$ = descriptor.template.src) !== null && _descriptor$template$ !== void 0 ? _descriptor$template$ : resourcePath;
    const idQuery = `&id=${id}`;
    const scopedQuery = hasScoped ? `&scoped=true` : ``;
    const srcQuery = descriptor.template.src ? `&src` : ``;
    const attrsQuery = attrsToQuery(descriptor.template.attrs);
    const query = `?vue&type=template${idQuery}${srcQuery}${scopedQuery}${attrsQuery}`;
    const templateRequest = `"${normalizePath(src) + query}"`;
    templateImport = `import { ${isServer ? "ssrRender" : "render"} } from ${templateRequest};`;
  }

  return templateImport;
}

function getScriptCode(descriptor, resourcePath) {
  let scriptImport = `const script = {}`;

  if (descriptor.script || descriptor.scriptSetup) {
    if (compilerSfc.compileScript) descriptor.script = compilerSfc.compileScript(descriptor);

    if (descriptor.script) {
      var _descriptor$script$sr;

      const src = (_descriptor$script$sr = descriptor.script.src) !== null && _descriptor$script$sr !== void 0 ? _descriptor$script$sr : resourcePath;
      const attrsQuery = attrsToQuery(descriptor.script.attrs, "js");
      const srcQuery = descriptor.script.src ? `&src` : ``;
      const query = `?vue&type=script${srcQuery}${attrsQuery}`;
      const scriptRequest = `"${normalizePath(src) + query}"`;
      scriptImport = `import script from ${scriptRequest};`;
      scriptImport += `\nexport * from ${scriptRequest};`;
    }
  }

  return scriptImport;
}

function getStyleCode(descriptor, resourcePath, id, preprocessStyles) {
  let stylesCode = ``;
  let hasCSSModules = false;

  if (descriptor.styles.length > 0) {
    descriptor.styles.forEach((style, i) => {
      var _style$src;

      const src = (_style$src = style.src) !== null && _style$src !== void 0 ? _style$src : resourcePath; // do not include module in default query, since we use it to indicate
      // that the module needs to export the modules json

      const attrsQuery = attrsToQuery(style.attrs, "css", preprocessStyles);
      const attrsQueryWithoutModule = attrsQuery.replace(/&module(=true|=[^&]+)?/, ""); // make sure to only pass id when necessary so that we don't inject
      // duplicate tags when multiple components import the same css file

      const idQuery = style.scoped ? `&id=${id}` : ``;
      const srcQuery = style.src ? `&src` : ``;
      const query = `?vue&type=style&index=${i}${srcQuery}${idQuery}`;
      const styleRequest = normalizePath(src) + query + attrsQuery;
      const styleRequestWithoutModule = normalizePath(src) + query + attrsQueryWithoutModule;

      if (style.module) {
        if (!hasCSSModules) {
          stylesCode += `\nconst cssModules = script.__cssModules = {};`;
          hasCSSModules = true;
        }

        stylesCode += genCSSModulesCode(i, styleRequest, styleRequestWithoutModule, style.module);
      } else {
        stylesCode += `\nimport "${styleRequest}";`;
      }
    });
  }

  return stylesCode;
}

function getCustomBlock(descriptor, resourcePath, filter) {
  const code = [];
  descriptor.customBlocks.forEach((block, index) => {
    var _block$src;

    if (!filter(block.type)) return;
    const src = (_block$src = block.src) !== null && _block$src !== void 0 ? _block$src : resourcePath;
    const srcQuery = block.src ? `&src` : ``;
    const attrsQuery = attrsToQuery(block.attrs, block.type);
    const query = `?vue&type=${block.type}&index=${index}${srcQuery}${attrsQuery}`;
    const request = `"${normalizePath(src) + query}"`;
    code.push(`import block${index} from ${request};`);
    code.push(`if (typeof block${index} === "function") { block${index}(script) };`);
  });
  return joinCode(code);
}

const createRollupError = (id, error) => {
  var _error$loc;

  return "code" in error ? {
    id,
    plugin: "vue3",
    pluginCode: String(error.code),
    message: error.message,
    frame: (_error$loc = error.loc) === null || _error$loc === void 0 ? void 0 : _error$loc.source,
    parserError: error,
    loc: error.loc && {
      file: id,
      line: error.loc.start.line,
      column: error.loc.start.column
    }
  } : {
    id,
    plugin: "vue3",
    message: error.message,
    parserError: error
  };
}; // these are built-in query parameters so should be ignored
// if the user happen to add them as attrs


const ignoreList = ["id", "index", "src", "type", "lang"];

function attrsToQuery(attrs, langFallback, forceLangFallback = false) {
  let query = ``;
  const attrsList = Object.entries(attrs).filter(([k]) => !ignoreList.includes(k));
  if (attrsList.length > 0) query += `&`;
  query += qs.stringify(Object.fromEntries(attrsList));

  if (attrs.lang && typeof attrs.lang === "string" && !forceLangFallback) {
    query += `&lang.${attrs.lang}`;
  } else if (langFallback) {
    query += `&lang.${langFallback}`;
  }

  return query;
}

const normalizeSourceMap = map => map ? { ...map,
  version: Number(map.version),
  mappings: typeof map.mappings === "string" ? map.mappings : ""
} : null;

function genCSSModulesCode(index, request, requestWithoutModule, moduleName) {
  const styleVar = `style${index}`;
  const code = [`import "${requestWithoutModule}";`, `import ${styleVar} from "${request}";`];
  const name = typeof moduleName === "string" ? moduleName : "$style";
  code.push(`cssModules["${name}"] = ${styleVar};`);
  return joinCode(code);
}

module.exports = index;
